package jess;

import java.io.Serializable;
import java.util.*;
import java.io.Writer;
import java.io.IOException;
import jess.factory.Factory;

/** **********************************************************************
 * The current list of facts.
 * <P>
 * (C) 1997 Ernest J. Friedman-Hill and Sandia National Laboratories
 * @author Ernest J. Friedman-Hill
 */


class FactList implements Serializable {

    private Map m_table = Collections.synchronizedMap(new HashMap());
    private ArrayList m_factsToAssert = new ArrayList(),
        m_factsToRetract = new ArrayList();
    private HashMap m_logicalSupport;
    private Token UNCONDITIONAL_SUPPORT;

    private int m_time = 0;
    
    FactList(Rete engine) {
        try { 
            UNCONDITIONAL_SUPPORT = 
                engine.getFactory().newToken(Fact.getNullFact(), RU.ADD);
        } catch (JessException je) {
            System.out.println("*** FATAL ***: Can't initialize Jess");
            je.printStackTrace();
            System.exit(0);
        }
    }

    void clear() {
        m_table.clear();
        m_time = 0;
        m_factsToAssert.clear();
        m_factsToRetract.clear();
        if (m_logicalSupport !=  null)
            m_logicalSupport.clear();
        m_nextFactId = 0;
    }

    int getTime() {
        return m_time;
    }

    void assignTime(Fact fact) {
        fact.updateTime(++m_time);
    }

    int doPreAssertionProcessing(Fact f) throws JessException {
        Object o = m_table.get(f);
        return (o == null) ? 0 : -1;
    }
  
    Fact findFactByFact(Fact f) throws JessException {
        return (Fact) m_table.get(f);
    }

    Fact findFactByID(int id) throws JessException {
        synchronized (m_table) {
            for (Iterator it = m_table.keySet().iterator(); it.hasNext();) {
                Fact f = (Fact) it.next();
                if (f.getFactId() == id)
                    return f;
            }
            return null;
        }
    }

    public void ppFacts(String name, Writer output) throws IOException {
        for (Iterator e = listFacts(); e.hasNext();) {
            Fact f = (Fact) e.next();
            if (name == null || f.getName().equals(name)) {
                output.write(f.toString());
                output.write("\n");
            }
        }
    }
    
    public void ppFacts(Writer output) throws IOException {
        ppFacts(null, output);
    }

    public Iterator listFacts() {        
        synchronized (m_table) {
            return new SortedIterator(m_table);
        }
    }

    /**
     * Successively incremented ID for asserted facts.
     */
    private Object m_idLock = "LOCK";
    private int m_nextFactId;

    private int nextFactId() {
        synchronized (m_idLock) {
            return m_nextFactId++;
        }
    }

    /**
     * Facts generated by LHS actions, not asserted until RHS time.
     */
  

    void setPendingFact(Fact fact, boolean assrt) {
        if (assrt)
            m_factsToAssert.add(fact);
        
        else {
            m_factsToRetract.add(fact);
        }
    }

    // Process any facts that were asserted by rule LHS processing -- i.e.,
    // for backwards chaining.
    void processPendingFacts(Rete r) throws JessException {
        synchronized (m_factsToAssert) {
            while (m_factsToAssert.size() > 0) {
                Fact f = (Fact) m_factsToAssert.get(0);
                m_factsToAssert.remove(0);
                _assert(f, true, r, r.getGlobalContext());
            }
        }
        
        synchronized (m_factsToRetract) {
            while (m_factsToRetract.size() > 0) {
                _retract((Fact) m_factsToRetract.get(0), r);
                m_factsToRetract.remove(0);
            }
        }
    }
    
    public Fact assertFact(Fact f, Rete r, Context context)
        throws JessException {
        Fact fact = _assert(f, true, r, context);
        if (fact != null)
            processPendingFacts(r);
        return fact;
    }

    Fact assertKeepID(Fact f, Rete r, Context context) throws JessException {
        Fact fact = _assert(f, false, r, context);
        if (fact != null)
            processPendingFacts(r);
        return fact;
    }

    private Fact _assert(Fact f, boolean newID, Rete r, Context context)
        throws JessException {
        synchronized (m_table) {

            if (context.getLogicalSupportNode() != null) {
                Token token = context.getToken();
                context.getLogicalSupportNode().dependsOn(f, context.getToken());
                addLogicalSupportFor(token, f   );
            } else if (m_logicalSupport != null) {
                addLogicalSupportFor(UNCONDITIONAL_SUPPORT, f);
            }

            // find any old copy
            if (r.doPreAssertionProcessing(f) != 0)
                return null;
        
            // insert the new fact
            if (newID)
                f.setFactId(nextFactId());

            r.broadcastEvent(JessEvent.FACT, f);        
                
            m_table.put(f, f);            
        
            // Send it to the Rete network
            r.processToken(RU.ADD, f);        

            return f;
        }
    }

    /**
     * Retract a fact. 
     * @param f A Fact object. Doesn't need to be the actual object
     * that appears on the fact-list; can just be a Fact that could
     * compare equal to one.
     * @exception JessException If anything goes wrong.
     */

    Fact retract(Fact f, Rete r) throws JessException {
        synchronized (m_table) {
            Fact ff;
            if ((ff = (Fact) m_table.get(f)) != null) {
                _retract(ff, r);
                processPendingFacts(r);
                return ff;
            }
            else
                return null;
        }
    }

    // f should be the actual fact
    private Fact _retract(Fact f, Rete r) throws JessException  {
        synchronized (m_table) {
            r.broadcastEvent(JessEvent.FACT | JessEvent.REMOVED, f);
            m_table.remove(f);
            r.processToken(RU.REMOVE, f);            
            return f;
        }
    }

    Fact _duplicate(ValueVector vv, Context context, Rete r)
        throws JessException {

        Value fv = vv.get(1).resolveValue(context);
        Fact fact;
        if (fv.type() == RU.INTEGER)
            fact = r.findFactByID(fv.intValue(context));
        else
            fact = (Fact) fv.externalAddressValue(context);
        
        if (fact == null || fact.getFactId() == -1)
            throw new JessException("duplicate", "no such fact", -1);
        else if (fact.isShadow()) 
            throw new JessException("duplicate",
                                    "Can't duplicate shadow fact",
                                    fact.toString());
            
        Deftemplate dt = fact.getDeftemplate();

        synchronized (r.getCompiler()) {

            fact = (Fact) fact.clone();
            
            for (int i= 2; i < vv.size(); i++) {
                
                // fetch the slot, value subexp, stored as a List 
                ValueVector svp = vv.get(i).listValue(context);
                String slotName = getSlotName(svp, context);
                int idx = dt.getSlotIndex(slotName);
                if (idx == -1)
                    throw new JessException("duplicate",
                                            "No such slot " + slotName +
                                            " in template",
                                            dt.getName());
                int type = dt.getSlotType(idx);
                
                // Set the value in the fact
                fact.setSlotValue(slotName,
                                  getSlotValue(svp, context, type));
            }

            // Add the new fact to the Rete network
            return assertFact(fact, r, context);
        }
    }

    // EJFH TODO this needs to be synchronized on m_table
    Fact _modify(ValueVector vv, Context context, Rete r)
        throws JessException {

        Value fv = vv.get(1).resolveValue(context);
        Fact fact;
        if (fv.type() == RU.INTEGER)
            fact = r.findFactByID(fv.intValue(context));
        else
            fact = (Fact) fv.externalAddressValue(context);
        
        if (fact == null || fact.getFactId() == -1)
            throw new JessException("modify", "no such fact", -1);

        Deftemplate dt = fact.getDeftemplate();

        //////////////////////////////////////////////////
        // First find out of this is a LHS object, not an ordinary fact
        // If so, call the appropriate mutators...
        //////////////////////////////////////////////////

        if (fact.isShadow()) {
            fact = modifyShadowFact(fact, vv, context, dt, r);        
            if (fact != null)
                processPendingFacts(r);
            return fact;
        }

        //////////////////////////////////////////////////
        // Not an object - just a fact.
        //////////////////////////////////////////////////

        // Remove the old fact from the Rete network
    
        synchronized (r.getCompiler()) {
            r.processToken(RU.REMOVE, fact);            
            m_table.remove(fact);
            
            for (int i= 2; i < vv.size(); i++) {
                
                // fetch the slot, value subexp, stored as a List 
                ValueVector svp = vv.get(i).listValue(context);
                String slotName = getSlotName(svp, context);
                int idx = dt.getSlotIndex(slotName);
                if (idx == -1)
                    throw new JessException("modify",
                                            "No such slot " + slotName +
                                            " in template",
                                            dt.getName());
                int type = dt.getSlotType(idx);
                
                // Set the value in the fact
                fact.setSlotValue(slotName, getSlotValue(svp, context, type));
            }

            // Add the modified fact back into the Rete network
            r.broadcastEvent(JessEvent.FACT | JessEvent.MODIFIED, fact);
            
            if (r.doPreAssertionProcessing(fact) == 0) {
                m_table.put(fact, fact);
                r.processToken(RU.ADD, fact);            
                if (fact != null)
                    processPendingFacts(r);
                return fact;
                
            } else {
                // EJFH Is this right, or should we just return?
                r.processToken(RU.REMOVE, fact);            
                if (fact != null)
                    processPendingFacts(r);
                return Fact.getNullFact();
            }    
        }
    }

    private Fact modifyShadowFact(Fact fact, ValueVector vv, Context context,
                                  Deftemplate dt, Rete r)
        throws JessException {

        Value ov = fact.getSlotValue("OBJECT");
        Funcall fc = new Funcall("set", r).
            arg(ov).arg(new Value("set", RU.STRING));
        fc.setLength(4);
        
        for (int i= 2; i < vv.size(); i++) {
            
            // fetch the slot, value subexp, stored as a List
            ValueVector svp = vv.get(i).listValue(context);
            String slotName = getSlotName(svp, context);
            
            fc.set(new Value(slotName, RU.ATOM), 2);
            int index = dt.getSlotIndex(slotName);
            if (index == -1)
                throw new JessException("modify",
                                        "No such slot " + slotName +
                                        " in template",
                                        dt.getName());
            
            int type = dt.getSlotType(index);
            fc.set(getSlotValue(svp, context, type), 3);
            
            fc.execute(r.getGlobalContext());
        }
        return fact;
    }


    private String getSlotName(ValueVector svp, Context c)
        throws JessException {
        
        return svp.get(0).stringValue(c);
    }
 
    private Value getSlotValue(ValueVector svp, Context c, int type)
        throws JessException {

        if (type == RU.SLOT) {
            Value v = svp.get(1).resolveValue(c);
            while (v.type() == RU.LIST)
                v = v.listValue(c).get(0).resolveValue(c);
        
            return v;

        } else { // MULTISLOT 
            ValueVector vv = new ValueVector();
            for (int i=1; i<svp.size(); i++) {
                Value listItem = svp.get(i).resolveValue(c);
                if (listItem.type() == RU.LIST) {
                    ValueVector sublist = listItem.listValue(c);
                    for (int j=0; j<sublist.size(); j++)
                        vv.add(sublist.get(j).resolveValue(c));
                }   
                else
                    vv.add(listItem);
            }
            return new Value(vv, RU.LIST);
        }
    }

    private void addLogicalSupportFor(Token token, Fact fact) {
        if (m_logicalSupport == null)
            m_logicalSupport = new HashMap();
        ArrayList list = (ArrayList) m_logicalSupport.get(fact);
        if (list == null) {
            list = new ArrayList();
            synchronized(m_logicalSupport) {
                m_logicalSupport.put(fact, list);
            }
        }
        list.add(token);
    }
    
    /**
     * If all the support for a fact disappears, remove it.
     */
    
    void removeLogicalSupportFrom(Token token, Fact fact) {
        if (m_logicalSupport == null)
            return;
        ArrayList list = (ArrayList) m_logicalSupport.get(fact);
        if (list == null) 
            return;
        synchronized(m_logicalSupport) {
            if (list.remove(token) && list.size() == 0) {
                m_logicalSupport.remove(fact);
                setPendingFact(fact, false);
            } 
        }
    }


}

class FactComparator implements Comparator, Serializable {
    public int compare(Object o1, Object o2) {
        return ((Fact) o1).getFactId() - ((Fact) o2).getFactId();
    }
}

class SortedIterator implements Iterator, Serializable {
    private Object[] m_arr;
    private int m_index;
    
    SortedIterator(Map map) {
        m_arr = map.keySet().toArray();
        Arrays.sort(m_arr, new FactComparator());
    }
    public boolean hasNext() {
        return m_index < m_arr.length;
    }
    public Object next() {
        return m_arr[m_index++];
    }
    public void remove() throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }
}



