package jess;

import java.util.*;
import java.io.*;

/** **********************************************************************
 * Node containing an arbitrary list of tests; used for TEST CE's and also
 * the base class for join nodes.
 ********************************************************************** */

class NodeTest extends Node implements Serializable
{
  private static Accelerator m_accelerator;
  private static boolean m_acceleratorChecked;
    
  
  /**
    The tests this node performs
    */

  int m_nTests = 0;
  Test[] m_tests = new Test[2];

  /**
   * Constructor
   */
  NodeTest(Rete engine) 
  {
  }

  void complete() throws JessException
  {
    // Nothing to do
  }

  /**
   * Bare NodeTests can only have Test1's in them.
   */
  void addTest(int test, int slot_idx, Value v) 
       throws JessException
  {
    addTest(test, slot_idx, -1, v); 
  }

  void addTest(int test, int slot_idx, int slot_sidx, Value v) 
       throws JessException
  {
    Test t = null;

    Funcall f = v.funcallValue(null);
    
    loadAccelerator();
    
    // if we have an accelerator, try to apply it
    if (m_accelerator != null)
      t = m_accelerator.speedup(f);
    
    // if no acceleration, use the standard Test1 class
    if (t == null)
      t = new Test1(test, slot_sidx, v);
    

    addTest(t);
  }

  void addTest(Test t)
  {
    if (m_nTests == m_tests.length)
      {
        Test[] temp = new Test[m_nTests*2];
        System.arraycopy(m_tests, 0, temp, 0, m_nTests);
        m_tests = temp;
      }

    m_tests[m_nTests++] = t;
  }

  /**
   * The classic 'you should never inherit from concrete classes' problem!
   */
  
  void addTest(int test, int token_idx, int left_idx, int leftSub_idx,
               int right_idx, int rightSub_idx)
       throws JessException
  {
    throw new JessException("NodeTest:addtest",
                            "Can't add Test2s to this class", "");
  }

  /**
   * For our purposes, two Node2's are equal if every test in
   * one has an equivalent test in the other, and if the test vectors are
   * the same size. The subclass NodeNot2 should never be shared, so we'll
   * report unequal always. This routine is used during network compilation,
   * not at runtime.
   */

  public boolean equals(Object o) 
  {
    if (this == o)
      return true;
    
    if (this.getClass() != o.getClass())
      return false;

    NodeTest n = (NodeTest) o;
      
    if (n instanceof NodeNot2 ||
        n.m_nTests != m_nTests)
      return false;

  outer_loop:
    for (int i=0; i<m_nTests; i++) 
      {
        Test t1 = m_tests[i];
        for (int j=0; j<m_nTests; j++) {
          if (t1.equals(n.m_tests[j]))
            continue outer_loop;
        }
        return false;
      }
    return true;
  }

  boolean callNodeRight(Token t) throws JessException
  {    
    if (t.m_tag == RU.CLEAR)
      return false;
    else
      return super.callNodeRight(t);
  }

  boolean callNodeLeft(Token token) throws JessException
  {
    broadcastEvent(JessEvent.RETE_TOKEN + LEFT, token); 
    if (token.m_tag == RU.CLEAR)
      {
        passAlong(token);
        return true;
      }
    
    int ntests = m_nTests;
    boolean result;

    if (token.m_tag == RU.REMOVE || ntests == 0)
       result = true;
    else
       {
          result = runTests(token, ntests);
          token = token.prepare(result);
       }

    if (result)
      {
        token = Rete.getFactory().newToken(token, Fact.getNullFact());
        token.updateTime(Rete.getEngine());
        passAlong(token);
      }
    return result;
  }

  boolean runTests(Token token, int ntests) throws JessException
  {
    try
      {
        Context context = new Context(Rete.getEngine().getGlobalContext());

        context.setToken(token);
        
        for (int i=0; i < ntests; i++) 
          {
            if (!m_tests[i].doTest(context))
              return false;
          }        
        return true;
      }
    catch (JessException re)
      {
        re.addContext("'test' CE");
        throw re;
      }
  }

    void loadAccelerator() throws JessException {
        // Try to load accelerator, if needed
        if (!m_acceleratorChecked) {
            m_acceleratorChecked = true;
            String classname;
            if ((classname = RU.getProperty("SPEEDUP")) != null) {
                try {
                    m_accelerator =
                        (Accelerator) Class.forName(classname).newInstance();
                } catch (Exception e) {
                    throw new JessException("NodeTest.addTest",
                                            "Can't load Accelerator class " +
                                            classname,
                                            e.getMessage());
                }
            }
        }
    }
    
  /**
    Two-input nodes always make calls to the left input of other nodes.
  */

  void passAlong(Token t) throws JessException
  { 
    Node [] sa = m_succ;
    for (int j=0; j<m_nSucc; j++) 
      {
        Node s = sa[j];
        s.callNodeLeft(t);
      }
  }

  /**
   * Describe myself
   * @return 
   */
  public String toString() 
  {
    StringBuffer sb = new StringBuffer(256);
    sb.append("[NodeTest ntests=");
    sb.append(m_nTests);
    sb.append(" ");
    for (int i=0; i<m_nTests; i++)
      {
        sb.append(m_tests[i].toString());
        sb.append(" ");
      }
    sb.append(";usecount = ");
    sb.append(m_usecount);
    sb.append("]");
    return sb.toString();
  }

  String getCompilationTraceToken() { return "2"; }

}





